from .base import Widget


class ScrollContainer(Widget):
    def __init__(self, id=None, style=None, horizontal=True, vertical=True):
        super().__init__(id=None, style=None)
        self.horizontal = horizontal
        self.vertical = vertical
        self._content = None

    @property
    def content(self):
        return self._content

    @content.setter
    def content(self, widget):
        self._content = widget
        self._content.window = self.window
        self._content.app = self.app
        self._set_content(widget)

    def _set_app(self, app):
        if self._content:
            self._content.app = self.app

    def _set_window(self, window):
        if self._content:
            self._content.window = self.window



from unittest import TestCase
if not hasattr(TestCase, 'assertIsNotNone'):
    # For Python2.6 compatibility
    from unittest2 import TestCase

from toga.constraint import Attribute, Constraint, InvalidConstraint


class AttributeTestCase(TestCase):
    def assertAttributeEqual(self, attr, multiplier, constant):
        self.assertEqual(attr.multiplier, multiplier)
        self.assertEqual(attr.constant, constant)

    def assertSimpleConstraintEqual(self, constraint, relation, multiplier, constant):
        self.assertEqual(constraint.relation, relation)
        self.assertAttributeEqual(constraint.attr, multiplier, constant)
        self.assertIsNone(constraint.related_attr)

    def assertRelatedConstraintEqual(self, constraint, relation, multiplier, constant, related_multiplier, related_constant):
        self.assertEqual(constraint.relation, relation)
        self.assertEqual(constraint.attr, multiplier, constant)
        self.assertEqual(constraint.related_attr, related_multiplier, related_constant)

    def test_add(self):
        "A constant can be added to an attribute"

        # A simple, empty attribute
        attr = Attribute(object(), Attribute.WIDTH)

        self.assertAttributeEqual(attr + 2, 1, 2)
        self.assertAttributeEqual(2 + attr, 1, 2)

        # A complex attribute
        attr = Attribute(object(), Attribute.WIDTH, 3, 7)

        self.assertAttributeEqual(attr + 2, 3, 9)
        self.assertAttributeEqual(2 + attr, 3, 9)

    def test_mult(self):
        "An attribute can be multiplied by a constant"

        # A simple, empty attribute
        attr = Attribute(object(), Attribute.WIDTH)

        self.assertAttributeEqual(attr * 2, 2, 0)
        self.assertAttributeEqual(2 * attr, 2, 0)

        # A complex attribute
        attr = Attribute(object(), Attribute.WIDTH, 3, 7)

        self.assertAttributeEqual(attr * 2, 6, 14)
        self.assertAttributeEqual(2 * attr, 6, 14)

    def test_div(self):
        "An attribute can be divided by a constant"

        # A simple, empty attribute
        attr = Attribute(object(), Attribute.WIDTH)

        self.assertAttributeEqual(attr / 2, 0.5, 0)

        # A complex attribute
        attr = Attribute(object(), Attribute.WIDTH, 3, 7)

        self.assertAttributeEqual(attr / 2, 1.5, 3.5)

    def test_sub(self):
        "A constant can be subtracted from an attribute"

        # A simple, empty attribute
        attr = Attribute(object(), Attribute.WIDTH)

        self.assertAttributeEqual(attr - 2, 1, -2)
        self.assertAttributeEqual(2 - attr, -1, 2)

        # A complex attribute
        attr = Attribute(object(), Attribute.WIDTH, 3, 5)

        self.assertAttributeEqual(attr - 2, 3, 3)
        self.assertAttributeEqual(2 - attr, -3, -3)

    def test_equality_constant(self):
        "When compared for equality with a constant, a single-attribute constraint is generated"

        self.assertSimpleConstraintEqual(
            Attribute(object(), Attribute.WIDTH) == 2,
            Constraint.EQUAL, 1, 2
        )

        self.assertSimpleConstraintEqual(
            3 * Attribute(object(), Attribute.WIDTH) + 7 == 9,
            Constraint.EQUAL, 3, 2
        )

        # Equality works both ways
        self.assertSimpleConstraintEqual(
            2 == Attribute(object(), Attribute.WIDTH),
            Constraint.EQUAL, 1, 2
        )

        self.assertSimpleConstraintEqual(
            9 == 3 * Attribute(object(), Attribute.WIDTH) + 7,
            Constraint.EQUAL, 3, 2
        )

    def test_less_than_constant(self):
        "When compared for less than or equal to a constant, a single-attribute constraint is generated"

        self.assertSimpleConstraintEqual(
            Attribute(object(), Attribute.WIDTH) <= 2,
            Constraint.LTE, 1, 2
        )

        self.assertSimpleConstraintEqual(
            3 * Attribute(object(), Attribute.WIDTH) + 7 <= 9,
            Constraint.LTE, 3, 2
        )

        # Operator works both ways
        self.assertSimpleConstraintEqual(
            2 >= Attribute(object(), Attribute.WIDTH),
            Constraint.LTE, 1, 2
        )

        self.assertSimpleConstraintEqual(
            9 >= 3 * Attribute(object(), Attribute.WIDTH) + 7,
            Constraint.LTE, 3, 2
        )

        # It's lazy, but the logic for < and <= are identical.
        self.assertSimpleConstraintEqual(
            3 * Attribute(object(), Attribute.WIDTH) + 7 < 9,
            Constraint.LTE, 3, 2
        )
        self.assertSimpleConstraintEqual(
            Attribute(object(), Attribute.WIDTH) < 2,
            Constraint.LTE, 1, 2
        )

    def test_greater_than_constant(self):
        "When compared for greater than or equal to a constant, a single-attribute constraint is generated"

        self.assertSimpleConstraintEqual(
            Attribute(object(), Attribute.WIDTH) >= 2,
            Constraint.GTE, 1, 2
        )

        self.assertSimpleConstraintEqual(
            3 * Attribute(object(), Attribute.WIDTH) + 7 >= 9,
            Constraint.GTE, 3, 2
        )

        # Operator works both ways
        self.assertSimpleConstraintEqual(
            2 <= Attribute(object(), Attribute.WIDTH),
            Constraint.GTE, 1, 2
        )

        self.assertSimpleConstraintEqual(
            9 <= 3 * Attribute(object(), Attribute.WIDTH) + 7,
            Constraint.GTE, 3, 2
        )

        # It's lazy, but the logic for > and >= are identical.
        self.assertSimpleConstraintEqual(
            3 * Attribute(object(), Attribute.WIDTH) + 7 > 9,
            Constraint.GTE, 3, 2
        )
        self.assertSimpleConstraintEqual(
            Attribute(object(), Attribute.WIDTH) > 2,
            Constraint.GTE, 1, 2
        )

    def test_equality_attribute(self):
        "When compared for equality with another attribute, a related constraint is generated"

        self.assertRelatedConstraintEqual(
            Attribute(object(), Attribute.WIDTH) == Attribute(object(), Attribute.WIDTH),
            Constraint.EQUAL, 1, 0, 1, 0
        )

        self.assertRelatedConstraintEqual(
            Attribute(object(), Attribute.WIDTH, 3, 7) == Attribute(object(), Attribute.WIDTH, 2, 5),
            Constraint.EQUAL, 3, 7, 2, 5
        )

    def test_less_than_attribute(self):
        "When compared for less than or equal to another attribute, a related constraint is generated"

        self.assertRelatedConstraintEqual(
            Attribute(object(), Attribute.WIDTH) <= Attribute(object(), Attribute.WIDTH),
            Constraint.LTE, 1, 0, 1, 0
        )

        self.assertRelatedConstraintEqual(
            Attribute(object(), Attribute.WIDTH, 3, 7) <= Attribute(object(), Attribute.WIDTH, 2, 5),
            Constraint.LTE, 3, 7, 2, 5
        )

        # It's lazy, but the logic for > and >= are identical.
        self.assertRelatedConstraintEqual(
            Attribute(object(), Attribute.WIDTH) < Attribute(object(), Attribute.WIDTH),
            Constraint.LTE, 1, 0, 1, 0
        )

        self.assertRelatedConstraintEqual(
            Attribute(object(), Attribute.WIDTH, 3, 7) < Attribute(object(), Attribute.WIDTH, 2, 5),
            Constraint.LTE, 3, 7, 2, 5
        )

    def test_greater_than_attribute(self):
        "When compared for greater than or equal to another attribute, a related constraint is generated"

        self.assertRelatedConstraintEqual(
            Attribute(object(), Attribute.WIDTH) >= Attribute(object(), Attribute.WIDTH),
            Constraint.GTE, 1, 0, 1, 0
        )

        self.assertRelatedConstraintEqual(
            Attribute(object(), Attribute.WIDTH, 3, 7) >= Attribute(object(), Attribute.WIDTH, 2, 5),
            Constraint.GTE, 3, 7, 2, 5
        )

        # It's lazy, but the logic for > and >= are identical.
        self.assertRelatedConstraintEqual(
            Attribute(object(), Attribute.WIDTH) > Attribute(object(), Attribute.WIDTH),
            Constraint.GTE, 1, 0, 1, 0
        )

        self.assertRelatedConstraintEqual(
            Attribute(object(), Attribute.WIDTH, 3, 7) > Attribute(object(), Attribute.WIDTH, 2, 5),
            Constraint.GTE, 3, 7, 2, 5
        )

    def test_invalid_comparisons(self):

        with self.assertRaises(InvalidConstraint):
            result = Attribute(object(), Attribute.WIDTH) == object()

        with self.assertRaises(InvalidConstraint):
            result = Attribute(object(), Attribute.WIDTH) > object()

        with self.assertRaises(InvalidConstraint):
            result = Attribute(object(), Attribute.WIDTH) >= object()

        with self.assertRaises(InvalidConstraint):
            result = Attribute(object(), Attribute.WIDTH) < object()

        with self.assertRaises(InvalidConstraint):
            result = Attribute(object(), Attribute.WIDTH) <= object()

        with self.assertRaises(InvalidConstraint):
            result = object() == Attribute(object(), Attribute.WIDTH)

        with self.assertRaises(InvalidConstraint):
            result = object() > Attribute(object(), Attribute.WIDTH)

        with self.assertRaises(InvalidConstraint):
            result = object() >= Attribute(object(), Attribute.WIDTH)

        with self.assertRaises(InvalidConstraint):
            result = object() < Attribute(object(), Attribute.WIDTH)

        with self.assertRaises(InvalidConstraint):
            result = object() <= Attribute(object(), Attribute.WIDTH)



import importlib
import os
import sys

from .constants import *

# Work around import loop issues (toga -> platform -> toga.interface) import
# all these things before we import the platform stuff
import toga.interface.app  # NOQA

__all__ = [
    '__version__',
    'platform'
]

# Examples of valid version strings
# __version__ = '1.2.3.dev1'  # Development release 1
# __version__ = '1.2.3a1'     # Alpha Release 1
# __version__ = '1.2.3b1'     # Beta Release 1
# __version__ = '1.2.3rc1'    # RC Release 1
# __version__ = '1.2.3'       # Final Release
# __version__ = '1.2.3.post1' # Post Release 1

__version__ = '0.2.0'

platform = None


def set_platform(module_name=None, locals=locals()):
    "Configures toga to use the specfied platform module"
    # Note - locals is deliberately passed in as an argument; because it is
    # a dictionary, this results in the module level locals dictionary being
    # bound as a local variable in this method -- and a persistent one,
    # because it was evaluated and bound at time of method import.

    # First check for an environment variable setting the platform
    if module_name is None:
        platform_name = os.environ.get('TOGA_PLATFORM')

        # If we don't have a manually defined platform, attempt to
        # autodetect and set the platform
        if platform_name is None:
            if sys.platform == 'ios':
                platform_name = 'iOS'
            elif sys.platform == 'tvos':
                platform_name = 'tvOS'
            elif sys.platform == 'watchos':
                platform_name = 'watchOS'
            elif sys.platform == 'android':
                platform_name = 'android'
            elif sys.platform == 'darwin':
                platform_name = 'cocoa'
            elif sys.platform in ('linux', 'linux2'):
                platform_name = 'gtk'
            elif sys.platform == 'win32':
                platform_name = 'win32'
            else:
                raise RuntimeError("Couldn't identify a supported host platform.")

        module_name = 'toga_' + platform_name

    # Purge any existing platform symbols in the toga module
    if locals['platform']:
        for symbol in locals['platform'].__all__:
            # Exclude __version__ from the list of symbols that is
            # ported, because toga itself has a __version__ identifier.
            if symbol != '__version__':
                locals.pop(symbol)

    # Import the new platform module
    try:
        # Set the new platform module into the module namespace
        locals['platform'] = importlib.import_module(module_name)

        # Export all the symbols *except* for __version__ from the platform module
        # The platform has it's own version identifier.
        locals.update(dict(
            (symbol, getattr(platform, symbol))
            for symbol in locals['platform'].__all__
            if symbol != '__version__'
        ))
    except ImportError as e:
        if e.name == module_name:
            locals['platform'] = None
            print("Couldn't import %s platform module; try running 'pip install %s'." % (module_name[5:], module_name))
            sys.exit(-1)
        else:
            raise

# On first import, do an autodetection of platform.
set_platform()



from .base import Widget


class SplitContainer(Widget):
    HORIZONTAL = False
    VERTICAL = True

    def __init__(self, id=None, style=None, direction=VERTICAL):
        super().__init__(id=None, style=None)
        self._content = None
        self.direction = direction

    @property
    def content(self):
        return self._content

    @content.setter
    def content(self, content):
        if len(content) < 2:
            raise ValueError('SplitContainer content must have at least 2 elements')

        self._content = content

        for widget in self._content:
            widget.window = self.window
            widget.app = self.app
            self._add_content(widget)

    def _set_app(self, app):
        if self._content:
            for content in self._content:
                content.app = self.app

    def _set_window(self, window):
        if self._content:
            for content in self._content:
                content.window = self.window



from .base import Widget


class Button(Widget):
    def __init__(self, label, id=None, style=None, on_press=None):
        super().__init__(id=id, style=style)
        self._label = label
        self.on_press = on_press

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, value):
        if value is None:
            self._label = ''
        else:
            self._label = str(value)
        self._set_label(value)

    def _set_label(self, value):
        raise NotImplementedError('Button widget must define _set_label()')



from colosseum import CSSNode


class Widget:
    def __init__(self, id=None, style=None):
        self._id = id
        self._parent = None
        self._children = None
        self._window = None
        self._app = None
        self.dirty = True
        self._layout_in_progress = False

        if style:
            self._style = style.apply(self)
        else:
            self._style = CSSNode(self)

    def __repr__(self):
        return "<%s:%s>" % (self.__class__.__name__, id(self))

    @property
    def id(self):
        return self._id

    @property
    def style(self):
        return self._style

    @property
    def parent(self):
        return self._parent

    @property
    def children(self):
        if self._children is None:
            return []
        else:
            return self._children

    def add(self, child):
        if self._children is None:
            raise ValueError('Widget cannot have children')
        self._children.append(child)

        child.app = self.app
        child._parent = self

        if self._parent:
            self._parent.dirty = True

        self._add_child(child)

    @property
    def app(self):
        return self._app

    @app.setter
    def app(self, app):
        if self._app:
            raise Exception("Widget %r is already associated with an App" % self)
        self._app = app
        self._set_app(app)
        if self._children is not None:
            for child in self._children:
                child.app = app

    @property
    def window(self):
        return self._window

    @window.setter
    def window(self, window):
        self._window = window
        self._set_window(window)
        if self._children is not None:
            for child in self._children:
                child.window = window

    def _update_layout(self, **style):
        """Force a layout update on the widget.

        The update request can be accompanied by additional style information
        (probably min_width, min_height, width or height) to control the
        layout.
        """
        # print("UPDATE LAYOUT OF ", self)
        if self._layout_in_progress:
            return
        self._layout_in_progress = True

        self.style.set(**style)

        # Recompute layout
        self.style.recompute()

        self._update_child_layout()

        # Set the frame for the widget to adhere to the new style.
        self._apply_layout()

        self._layout_in_progress = False

    def _update_child_layout(self):
        # print("UPDATE CHILD LAYOUT - widget")
        if self._children is not None:
            for child in self.children:
                # if child.is_container:
                child._update_layout()



class App(object):
    '''Toga App

    The App is the top level of any GUI program. It is the manager of all the
    other bits of the GUI app: the main window and events that window generates
    like user input.

    When you create an App you need to provide it a name, an id for uniqueness
    (by convention, the identifier is a "reversed domain name".) and an
    optional startup function which should run once the App has initialised.
    The startup function typically constructs some initial user interface.

    Once the app is created you should invoke the main_loop() method, which
    will hand over execution of your program to Toga to make the App interface
    do its thing.

    Here is the absolute minimum App::

        app = toga.App('Empty App', 'org.pybee.empty')
        app.main_loop()
    '''
    def __init__(self, name, app_id, icon=None, startup=None, document_types=None):
        self.name = name
        self.app_id = app_id

        self.icon = icon

        self.document_types = document_types
        self._documents = []

        self._startup_method = startup

    @property
    def documents(self):
        '''Return the list of documents associated with this app.'''
        return self._documents

    def add_document(self, doc):
        '''Add a new document to this app.'''
        doc.app = self
        self._documents.append(doc)

    def open_document(self, fileURL):
        '''Add a new document to this app.'''
        raise NotImplementedError('Application class must define open_document()')

    def main_loop(self):
        '''Invoke the application to handle user input.

        This method typically only returns once the application is exiting.
        '''
        raise NotImplementedError('Application class must define main_loop()')



from .base import Widget


class ProgressBar(Widget):
    def __init__(self, id=None, style=None, max=None, value=None):
        super().__init__(id=None, style=None)
        self._value = value
        self._running = False



from .base import Widget


class ImageView(Widget):
    def __init__(self, id=None, style=None, image=None):
        super().__init__(id=None, style=None, image=None)

    # @property
    # def alignment(self):
    #     return self._alignment

    # @alignment.setter
    # def alignment(self, value):
    #     self._alignment = value
    #     self._impl.setAlignment_(NSTextAlignment(self._alignment))

    # @property
    # def scaling(self):
    #     return self._scaling

    # @scaling.setter
    # def scaling(self, value):
    #     self._scaling = value
    #     self._impl.setAlignment_(NSTextAlignment(self._scaling))
